/**
@author Cesar Vasquez Ibaceta
@version 1.0 13-04-2004
*/

import javax.swing.*;
import java.lang.String.*;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import java.awt.Toolkit.*;
import java.awt.geom.*;
import java.awt.Graphics2D.*;

/**
	Esta clase, nos permite crear nuestro panel de dibujo, en donde pondremos
	las figuras que dibujemos.
	Tambien se encarga de crear nuestras figuras, que seran objetos del tipo
	Line2D, Rectangle2D o Ellipse2D segun sea el caso.
	Ademas es la encargada de implementar el MouseListener y MouseMoveListener
	para estar atenta a cuando queramos dibujar.
*/
public class drawPanel extends JPanel
{
	private Point2D P1;
	private Point2D P2;

	private ArrayList elipses;
	private ArrayList lineas;
	private ArrayList rectangulos;

	private Ellipse2D E;
	private Rectangle2D R;
	private Line2D L;

	private String option;
	private double dx, dy, d1, d2;
	private double theta, alpha, gamma;
	private int counter;
	private double H;
	private double W;
	private double zoom, zoomOld;

	/**
		Crea un un Panel en donde dibujaremos las figuras.
		Ademas, se encarga de inicializar todas las variables criticas para que la
		funcion paintComponent no reciba arreglos y punteros vacios.
	*/
	public drawPanel()
	{
		elipses =  new ArrayList();
		lineas =  new ArrayList();
		rectangulos =  new ArrayList();
		option = "NULL";

		setUp();			//se inicializan las variables

		addMouseMotionListener(new MouseMotionHandler());
		addMouseListener(new MouseHandler());
		setBackground(Color.WHITE);
	}



	/**
		Metodo que inicializa casi todas las variables del Objeto.
	*/
	private void setUp( )
	{
		dx = 0;
		dy = 0;
		d1 = 0;
		d2 = 0;
		E = null;
		R = null;
		L = null;
		theta = 0;
		alpha = 0;
		gamma = 0;
		zoom = 1;
		zoomOld = 1;
		counter = 0;

		P1 = new Point2D.Double(0,0);
		P2 = new Point2D.Double(0,0);

	}




	/**
		Recibe un llamado cuando es presionado algun boton. Ademas se usa para recibir
		el nombre del boton presionado como parametro
		@param mensaje recibe el nombre de un boton en forma de string cuando es presionado
		algun boton.
	*/
	public void parametro(String mensaje)
	{
		option = mensaje;				//se recibe que boton fue apretado
		if(option.equals("limpiar")) limpiar();	//si se presiono "limpiar", se limpia
		setUp();		//se limpian las variables
	}


	/**
		Limpia los arreglos de figura, para poder limpiar el area de dibujo.
		Es llamada por el boton "limpiar" mediante un ActionListener.
	*/
	public void limpiar()
	{
		// se vacian todos los arreglos.
		rectangulos.clear();
		elipses.clear();
		lineas.clear();
		repaint();
	}


	/**
		Este metodo se encarga de entregar el punto que se encuentra en la esquina superior
		izquierda del ractangulo formado por los puntos p1 y p2. Ademas setea el valor del
		ancho y la altura del mismo.
		@param p1 es punto junto con p2 sirve para determinar la altura y el ancho del
		rectangulo formado por ellos ademas del punto en la esquina superior izquierda.
		@param p2 es punto junto con p1 sirve para determinar la altura y el ancho del
		rectangulo formado por ellos ademas del punto en la esquina superior izquierda.
		@return punto retorna el punto de la esquina superior izquierda.
	*/
	private Point2D direccion(Point2D p1, Point2D p2)
	{
		Point2D punto;
		double x;
		double y;
		double ancho;
		double alto;

		ancho = p1.getX() - p2.getX();
		alto = p1.getY() - p2.getY();

		if( ancho > 0 ) x = p2.getX();
		else x = p1.getX();

		if( alto > 0 ) y = p2.getY();
		else y = p1.getY();

		W = Math.abs( ancho );
		H = Math.abs( alto );
		punto = new Point2D.Double( x , y );
		return punto;
	}



	/**
		Crea un objeto Elipse a partir de dos puntos.
		@param p1 punto inicial, cuando el boton del mouse es presionado.
		@param p2 punto final, cuando el boton del mouse es soltado.
		@return forma devuelve un objeto Ellipse2D
	*/
	private Ellipse2D Elipse(Point2D p1, Point2D p2)
	{
		Point2D punto;

		punto = direccion(p1, p2);
		Ellipse2D forma = new Ellipse2D.Double(punto.getX(), punto.getY(), W, H);
		return forma;		//se crea un objeto "Ellipse2D" y se devuelve.
	}



	/**
		Crea un objeto Rectangulo a partir de dos puntos.
		@param p1 punto inicial, cuando el boton del mouse es presionado.
		@param p2 punto final, cuando el boton del mouse es soltado.
		@return forma devuelve un objeto Rectangle2D.
	*/
	private Rectangle2D Rectangulo(Point2D p1, Point2D p2)
	{
		Point2D punto;

		punto = direccion(p1, p2);
		Rectangle2D forma = new Rectangle2D.Double(punto.getX(), punto.getY(), W, H);
		return forma;		//se crea un objeto "Rectangle2D" y se devuelve.
	}



	/**
		Crea un objeto Linea a partir de dos puntos.
		@param p1 punto inicial, cuando el boton del mouse es presionado.
		@param p2 punto final, cuando el boton del mouse es soltado.
		@return forma devuelve un objeto Line2D.
	*/
	private Line2D Linea(Point2D p1, Point2D p2)
	{
		Point2D punto1 = new Point2D.Double(p1.getX(), p1.getY());
		Point2D punto2 = new Point2D.Double(p2.getX(), p2.getY());
		Line2D forma = new Line2D.Double(punto1, punto2);
		return forma;		//se crea un objeto "Line2D" y se devuelve.
	}


	/**
		Metodo encargado de pintar y repintar lo que dibujemos
		@param g objeto grafico que representa todo el panel en el que nos encontramos.
	*/
	public void paintComponent(Graphics g)
	{
		int i, k , j;
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;

		//se rota, mueve, o escala solo si se presionan los botones correspondientes
		if( option.equals("mover") ) g2.translate(dx, dy);
		else if( option.equals("escalar") ) g2.scale(zoom, zoom);
		else if( option.equals("rotar") ) g2.rotate(theta, P1.getX(), P1.getY());

		//se dibujan las figuras dentro de los arreglos.
		for(i = 0; i < rectangulos.size(); i++ )
			g2.draw((Rectangle2D)rectangulos.get(i));
		for(k = 0; k < elipses.size(); k++ )
			g2.draw((Ellipse2D)elipses.get(k));
		for(j = 0; j < lineas.size(); j++ )
			g2.draw((Line2D)lineas.get(j));

		//se dibuja la figura que se esta dibujando en este momento (se refresca)
		if( R != null ) g2.draw((Rectangle2D)R);
		else if( E != null ) g2.draw((Ellipse2D)E);
		else if( L != null ) g2.draw((Line2D)L);
	}


	/**
		Metodo encargado de implementar el escalamiento de nuestras figuras.
		Es la encargada de escribir en la variable global "zoom" el factor de
		escalamiento asociado al arrastre del mouse.
		@param event este metodo es llamado solo desde algun Listener, por lo
		tanto, este parametro representa el MouseEvent asociado a ese Listener
		@param call recibe un String que indica que MouseListener esta llamando
		a este metodo.
	*/
	private void escalar(MouseEvent event, String call)
	{
		Point2D P3;

		/* el metodo en que se escala, es calcular la distancia entre el primer y el ultimo punto
		y dividirla por la distancia entre el primer y segundo punto. este es el factor de
		escalamiento "zoom"
		*/
		if(call.equals("click"))
		{

			if(counter == 0) P1 = event.getPoint();
			else P2 = event.getPoint();
			counter++;
		}

		if(call.equals("release"))
		{
			if(!(counter == 2)) return;
			counter = 0;
			zoomOld = zoom;		//necesario para mantener el zoom mientras se esta en modo escalar.
		}

		if(call.equals("drag"))
		{
			if(counter == 2)
			{
				P3 = event.getPoint();
				if(P1.distance(P3) == 0) return;		//comprobaciones para evitar divisiones
				else if(P1.distance(P2) == 0) return;	//por cero,	y punteros vacios.

				zoom = zoomOld*(P1.distance(P3)/P1.distance(P2));
				repaint();

			}
		}
	}


	/**
		Metodo encargado de implementar la rotacion de nuestras figuras.
		Es la encargada de escribir en la variable global "theta", el algulo
		de rotacion asociado a arraste del Mouse.
		@param event este metodo es llamado solo desde algun Listener, por lo
		tanto, este parametro representa el MouseEvent asociado a ese Listener
		@param call recibe un String que indica que MouseListener esta llamando
		a este metodo.
	*/
	private void rotar(MouseEvent event, String call)
	{
		Point2D P3;

		/* el modo de rotar es calcular el angulo entre el tercer punto y el primero
			para ello se usa un punto auxiliar (P2) y se utiliza el teorema del coseno. */

		if(call.equals("click"))
		{
			if(counter == 0) P1 = event.getPoint();
			else
			{
				P2 = new Point2D.Double(P1.getX() + 10, P1.getY());
				P3 = event.getPoint();
				gamma = Angulo(P1, P2, P3);
			}
			counter++;
		}
		else if(call.equals("release"))
		{
			if(!(counter == 2)) return;
			setUp();
		}
		else if(call.equals("drag"))
		{
			if(counter == 2)
			{
				P3 = event.getPoint();
				theta =  gamma + Angulo(P1, P2, P3);
				repaint();
			}
			else counter = 0;
		}
	}


	/**
		Funcion encargada de calcular el angulo de rotacion a partir de dos puntos y
		el arrastre (Pc), mediante el teorema del coseno.
		@param beta angulo de rotacion a partir de los puntos entregados. en radianes.
	*/
	private double Angulo(Point2D Pa, Point2D Pb, Point2D Pc)
	{
		double a, b, c;
		double beta;

		a = Pc.distance(Pb);
		b = Pa.distance(Pb);
		c = Pc.distance(Pa);

		if(b == 0 && c == 0) return 0;

		double temp = (- Math.pow(a,2) + Math.pow(b,2) + Math.pow(c,2)) / (2*b*c) ;

		if( Math.abs(temp) > 1 ) return 0;
		beta = Math.acos(temp);

		if( Pa.getY() > Pc.getY()) return (- beta);

		return beta;
	}


	/**
		Clase Listener del mouse para nuestro panel de dibujo, detecta alguna acciones
		estaticas del mouse, como clicks y o numero de clicks etc.
	*/
	private class MouseHandler extends MouseAdapter
	{
		/**
			Reacciona cuando el boton del mouse es presionado.
			@param event Entrega info sobre el evento. En nuestro caso nos interesa el lugar
			en donde se produjo el <i>click</i>
		*/
		public void mousePressed(MouseEvent event)
		{
			//en el modo rotar y escalar, se gurada P1 en otro metodo.
			if(!option.equals("rotar") && !option.equals("escalar")) P1 = event.getPoint();

			//"click" string para indicar al metodo rotar y escalar, que el mouse fue presionado.
			else if(option.equals("rotar")) rotar(event, "click");
			else if(option.equals("escalar")) escalar(event, "click");
		}

		/**
			Reacciona cuando el boton del mouse es soltado.
			@param event Entrega info sobre el evento. En este caso, nos interesa el lugar
			en donde el boton del mouse fue soltado.
		*/
		public void mouseReleased(MouseEvent event)
		{
			if(!option.equals("rotar") && !option.equals("escalar"))
			{
				P2 = event.getPoint();

				if(option.equals("ELIPSE"))
				{
					E = Elipse(P1, P2);
					elipses.add(E);
				}
				else if(option.equals("RECTANGULO"))
				{
					R = Rectangulo(P1, P2);
					rectangulos.add(R);
				}
				else if(option.equals("LINEA"))
				{
					L = Linea(P1, P2);
					lineas.add(L);
				}
				else if(option.equals("mover") )
				{
					d1 = dx;
					d2 = dy;
				}
			R = null;
			L = null;
			E = null;

			repaint();
			}
			//"release" string para indicar al metodo rotar y escalar, que el mouse fue soltado.
			if(option.equals("rotar")) rotar(event, "release");
			else if(option.equals("escalar")) escalar(event, "release");


		}
	}



	/**
		Listener que detecta eventos activos de nuestro mouse, tales como: movimiento, arrastre,
		etc. y produce alguna accion.
	*/
	private class MouseMotionHandler extends MouseMotionAdapter
	{
		private Point2D P3;

		/**
			Reacciona cuando el mouse es arrastardo, arrojando informacion sobre el ultimo punto.
			@param event entrega informacion sobre el evento. En este caso nos interesa saber
			cual fue el utlimo punto en que el mouse fue arrastrado.
		*/
		public void mouseDragged(MouseEvent event)
		{
			if(!option.equals("rotar") && !option.equals("escalar")) P2 = event.getPoint();

			if(option.equals("ELIPSE"))
			{
				E = Elipse(P1, P2);
				repaint();
			}
			else if(option.equals("RECTANGULO"))
			{
				R = Rectangulo(P1, P2);
				repaint();
			}
			else if(option.equals("LINEA"))
			{
				L = Linea(P1, P2);
				repaint();
			}
			else if(option.equals("mover"))
			{
				dx = event.getX() - P1.getX();
				dy = event.getY() - P1.getY();
				dx = d1 + dx;
				dy = d2 + dy;
				repaint();
			}

			//"drag" string para indicar al metodo rotar y escalar, que el mouse fue arrastrado.
			else if(option.equals("rotar")) rotar(event, "drag");
			else if(option.equals("escalar")) escalar(event, "drag");
		}
	}
}

